#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "config.h"

#ifndef HAVE_GETOPT
#include "getopt.c"
#endif

#define BLOCKSIZE 1024

int convertnsline(FILE * fileout, char *line, int force);

int main(int argc, char *argv[])
{
    int c;
    int pcorrupt = 0, force = 0;
    long int size = 0, slide = 0, corrupt = 0;
    char *line = NULL;
    char *outfile = NULL;
    int opt;
    extern int optind;
    extern char *optarg;
    FILE *fileout;
    FILE *filein;

    while ((opt = getopt(argc, argv, "o:hvcf")) != EOF) {
	switch (opt) {
	case 'o':
	    outfile = optarg;
	    break;

	case 'h':
	    printf("nstoelf [-o output-file] [-c] [-f] [files] ...\n\n-o output-file  File to write output to. Default is stdout.\n-c              Write corrupt lines to stderr.\n-f              Force conversion of corrupt lines. Garbage in, garbage out.\n");
	    exit(0);

	case 'v':
	    printf("nstoelf - " PACKAGE " " VERSION "\n");
	    exit(0);

	case 'c':
	    pcorrupt = 1;
	    break;

	case 'f':
	    force = 1;
	    break;

	}
    }

    /* open up file for output */
    if (outfile) {
	if (!(fileout = fopen(outfile, "w"))) {
	    fprintf(stderr, "File open error\n");
	    exit(1);
	}
    }
    else
	fileout = stdout;

    /* open up and read each file for input */
    opt = optind;
    if (opt == argc)
	filein = stdin;
    else
	filein = NULL;

    /* loop through each file */
    while (filein || opt < argc) {

	/* open the files for reading */
	if (!filein) {
	    if (!strcmp(argv[opt], "-"))
		filein = stdin;
	    else if (!(filein = fopen(argv[opt], "r"))) {
		perror(argv[opt]);
		exit(1);
	    }
	}

	/* get each line of text until EOF */
	while ((c = fgetc(filein)) != EOF) {
	    /* enough space in buffer? no, make some more */
	    if (slide >= size) {
		size += BLOCKSIZE;
		if (!(line = (char *) realloc(line, size))) {
		    fprintf(stderr, "Memory allocation error\n");
		    exit(1);
		}
	    }

	    /* save character to buffer */
	    line[slide++] = (char) c;

	    /* reached an end of line? if so, deal with the line */
	    if (c == 10) {

		/* the line may be terminated with a CR, if so chop it off
		 * otherwise replace the LF with a NULL to indicate end of string */
		if ((slide > 0) && (line[slide - 1] == 13))
		    line[slide - 2] = 0;
		else
		    line[slide - 1] = 0;

		/* is the line a comment? if so, ignore it */
		switch (line[0]) {
		case 0:
		case '\t':
		case ' ':
		case ';':
		case '!':
		case '{':{
			break;
		    }
		default:{
			if (convertnsline(fileout, line, force)) {
			    if (pcorrupt)
				fprintf(stderr, "%s\n", line);
			    corrupt++;
			}
		    }
		}

		free(line);
		line = NULL;
		slide = 0;
		size = 0;
	    }

	}
	if (line)
	    free(line);

	/* close file (but only if the file was a real file) and loop */
	if (filein != stdin)
	    fclose(filein);
	filein = NULL;
	opt++;
    }

    /* close out output file */
    if (fileout != stdout)
	fclose(fileout);

    /* report corrupt line count */
    if (corrupt && !pcorrupt)
	fprintf(stderr, "%ld corrupt/incomplete lines ignored.\n", corrupt);

    exit(0);
}

/* get field number in a line */
/* buffer returned is size + 127 bytes so that it may be overwritten by the calling routine */
char *getfield(char *line, int field)
{
    char *buf;
    int start = 0, end = 0, slide = 0;
    int ws = 0;
    int br = 0;

    /* if ws is 0, we are finding text. if ws is not 0, we are finding whitespace */
    /* if br is not 0, we are inside a set of brackets */
    while (field && line[slide] && (line[slide] != 13)) {

	/* deal with brackets */
	if (line[slide] == '"') {
	    if (br)
		br = 0;
	    else
		br = 1;
	}

	/* find whitespace */
	switch (line[slide]) {
	case 8:
	case ' ':{
		if (!br && ws) {
		    ws = 0;
		}
		break;
	    }
	default:{
		if (!ws) {
		    field--;
		    ws = 1;
		    start = slide;
		}
	    }
	}
	slide++;
    }

    /* we found the start, now find the end */
    while (!end && line[slide] && (line[slide] != 13)) {

	/* find whitespace */
	switch (line[slide]) {
	case '"':{
		if (br)
		    br = 0;
		else
		    br = 1;
		break;
	    }
	case 8:
	case ' ':{
		if (!br) {
		    end = slide;
		}
		break;
	    }
	}
	slide++;
    }

    /* if !end, we found the last field */
    if (!end)
	end = slide;

    /* if field not found, return "" */
    if (field)
	start = end = 0;

    /* create a buffer for the field and return it - buffer contains an extra 127 bytes */
    if (!(buf = (char *) malloc(end - start + 128)))
	return (NULL);
    strncpy(buf, line + start, end - start);
    buf[end - start] = 0;
    return (buf);
}

void plustospace(char *str)
{
    register int x;

    for (x = 0; str[x]; x++)
	if (str[x] == '+')
	    str[x] = ' ';
}

#define NS_CLIENTIP "%Ses->client.ip%"
#define NS_AUTHUSER1 "%Req->vars.auth-user%"
#define NS_AUTHUSER2 "%Req->vars.pauth-user%"
#define NS_SYSDATE "%SYSDATE%"
#define NS_REQUEST1 "%Req->reqpb.clf-request%"
#define NS_REQUEST2 "%Req->reqpb.proxy-request%"
#define NS_STATUS "%Req->srvhdrs.clf-status%"
#define NS_CONTENTLENGTH1 "%Req->srvhdrs.content-length%"
#define NS_CONTENTLENGTH2 "%Req->headers.content-length%"
#define NS_CONTENTLENGTH3 "%Req->vars.p2c-cl%"
#define NS_REFERRER "%Req->headers.referer%"
#define NS_USERAGENT "%Req->headers.user-agent%"

#define DEFAULT_FORMAT "%[^ ] - %[^ ] [%[^]]] \"%[^\"]\" %[^ ] %[^ ] \"%[^\"]\" \"%[^\"]\""

/* convert the line as appropriate */
int sysdateno = 3;
int ipno = 1;
int remoteuserno = 0;
int authuserno = 2;
int requestno = 4;
int statusno = 5;
int contentlengthno = 6;
int useragentno = 8;
int referrerno = 7;
char *formatline = NULL;
int formatparms = 8;
char *vp[10]; /* 8 options + 1 NULL plus one spare */

int convertnsline(FILE * fileout, char *line, int force)
{
    int x, y, z, length;
    int bypass = 0;
    char *remoteuser;
    char *authuser;

    /* make space for new format line */
    if (!formatline) {
	if (!(formatline=(char *)malloc(sizeof(DEFAULT_FORMAT)+1))) {
	    fprintf(stderr, "Memory allocation error\n");
	    exit(1);
	}
	strcpy(formatline, DEFAULT_FORMAT);
    }

    /* check if the line describes a line format - if so, parse that format
     * and then ignore that line */
    if (('f' == line[0])&&(!strncmp(line,"format=", 7))) {

	/* reset defaults to blank */
	sysdateno = 0;
	ipno = 0;
	authuserno = 0;
	remoteuserno = 0;
	requestno = 0;
	statusno = 0;
	contentlengthno = 0;
	useragentno = 0;
	referrerno = 0;

	/* make space for new format line */
	free(formatline);
	if (!(formatline=(char *)malloc(strlen(line)+1024))) {
	    fprintf(stderr, "Memory allocation error\n");
	    exit(1);
	}

	/* extract format string from line */
	length = strlen(line);
	x = 7;
	y = 0;
	z = 1;
	while (x < length) {
	    if (line[x] == '%') {
		if (!strncmp(line+x, NS_CLIENTIP, sizeof(NS_CLIENTIP)-1)) {
		    if (ipno) {
			sprintf(formatline+y, "%%*[^%c]",line[x+sizeof(NS_CLIENTIP)-1]);
			y += 6;
		    }
		    else {
			sprintf(formatline+y, "%%[^%c]",line[x+sizeof(NS_CLIENTIP)-1]);
			y += 5;
			ipno = z++;
		    }
		    x += sizeof(NS_CLIENTIP)-1;
		}
		else if (!strncmp(line+x, NS_AUTHUSER1, sizeof(NS_AUTHUSER1)-1)) {
		    if (authuserno) {
			sprintf(formatline+y, "%%*[^%c]",line[x+sizeof(NS_AUTHUSER1)-1]);
			y += 6;
		    }
		    else {
			sprintf(formatline+y, "%%[^%c]",line[x+sizeof(NS_AUTHUSER1)-1]);
			y += 5;
			authuserno = z++;
		    }
		    x += sizeof(NS_AUTHUSER1)-1;
		}
		else if (!strncmp(line+x, NS_AUTHUSER2, sizeof(NS_AUTHUSER2)-1)) {
		    if (authuserno) {
			sprintf(formatline+y, "%%*[^%c]",line[x+sizeof(NS_AUTHUSER2)-1]);
			y += 6;
		    }
		    else {
			sprintf(formatline+y, "%%[^%c]",line[x+sizeof(NS_AUTHUSER2)-1]);
			y += 5;
			authuserno = z++;
		    }
		    x += sizeof(NS_AUTHUSER2)-1;
		}
		else if (!strncmp(line+x, NS_SYSDATE, sizeof(NS_SYSDATE)-1)) {
		    if (sysdateno) {
			sprintf(formatline+y, "%%*[^%c]",line[x+sizeof(NS_SYSDATE)-1]);
			y += 6;
		    }
		    else {
			sprintf(formatline+y, "%%[^%c]",line[x+sizeof(NS_SYSDATE)-1]);
			y += 5;
			sysdateno = z++;
		    }
		    x += sizeof(NS_SYSDATE)-1;
		}
		else if (!strncmp(line+x, NS_REQUEST1, sizeof(NS_REQUEST1)-1)) {
		    if (requestno) {
			sprintf(formatline+y, "%%*[^%c]",line[x+sizeof(NS_REQUEST1)-1]);
			y += 6;
		    }
		    else {
			sprintf(formatline+y, "%%[^%c]",line[x+sizeof(NS_REQUEST1)-1]);
			y += 5;
			requestno = z++;
		    }
		    x += sizeof(NS_REQUEST1)-1;
		}
		else if (!strncmp(line+x, NS_REQUEST2, sizeof(NS_REQUEST2)-1)) {
		    if (requestno) {
			sprintf(formatline+y, "%%*[^%c]",line[x+sizeof(NS_REQUEST2)-1]);
			y += 6;
		    }
		    else {
			sprintf(formatline+y, "%%[^%c]",line[x+sizeof(NS_REQUEST2)-1]);
			y += 5;
			requestno = z++;
		    }
		    x += sizeof(NS_REQUEST2)-1;
		}
		else if (!strncmp(line+x, NS_STATUS, sizeof(NS_STATUS)-1)) {
		    if (statusno) {
			sprintf(formatline+y, "%%*[^%c]",line[x+sizeof(NS_STATUS)-1]);
			y += 6;
		    }
		    else {
			sprintf(formatline+y, "%%[^%c]",line[x+sizeof(NS_STATUS)-1]);
			y += 5;
			statusno = z++;
		    }
		    x += sizeof(NS_STATUS)-1;
		}
		else if (!strncmp(line+x, NS_CONTENTLENGTH1, sizeof(NS_CONTENTLENGTH1)-1)) {
		    if (contentlengthno) {
			sprintf(formatline+y, "%%*[^%c]",line[x+sizeof(NS_CONTENTLENGTH1)-1]);
			y += 6;
		    }
		    else {
			sprintf(formatline+y, "%%[^%c]",line[x+sizeof(NS_CONTENTLENGTH1)-1]);
			y += 5;
			contentlengthno = z++;
		    }
		    x += sizeof(NS_CONTENTLENGTH1)-1;
		}
		else if (!strncmp(line+x, NS_CONTENTLENGTH2, sizeof(NS_CONTENTLENGTH2)-1)) {
		    if (contentlengthno) {
			sprintf(formatline+y, "%%*[^%c]",line[x+sizeof(NS_CONTENTLENGTH2)-1]);
			y += 6;
		    }
		    else {
			sprintf(formatline+y, "%%[^%c]",line[x+sizeof(NS_CONTENTLENGTH2)-1]);
			y += 5;
			contentlengthno = z++;
		    }
		    x += sizeof(NS_CONTENTLENGTH2)-1;
		}
		else if (!strncmp(line+x, NS_CONTENTLENGTH3, sizeof(NS_CONTENTLENGTH3)-1)) {
		    if (contentlengthno) {
			sprintf(formatline+y, "%%*[^%c]",line[x+sizeof(NS_CONTENTLENGTH3)-1]);
			y += 6;
		    }
		    else {
			sprintf(formatline+y, "%%[^%c]",line[x+sizeof(NS_CONTENTLENGTH3)-1]);
			y += 5;
			contentlengthno = z++;
		    }
		    x += sizeof(NS_CONTENTLENGTH3)-1;
		}
		else if (!strncmp(line+x, NS_REFERRER, sizeof(NS_REFERRER)-1)) {
		    if (referrerno) {
			sprintf(formatline+y, "%%*[^%c]",line[x+sizeof(NS_REFERRER)-1]);
			y += 6;
		    }
		    else {
			sprintf(formatline+y, "%%[^%c]",line[x+sizeof(NS_REFERRER)-1]);
			y += 5;
			referrerno = z++;
		    }
		    x += sizeof(NS_REFERRER)-1;
		}
		else if (!strncmp(line+x, NS_USERAGENT, sizeof(NS_USERAGENT)-1)) {
		    if (useragentno) {
			sprintf(formatline+y, "%%*[^%c]",line[x+sizeof(NS_USERAGENT)-1]);
			y += 6;
		    }
		    else {
			sprintf(formatline+y, "%%[^%c]",line[x+sizeof(NS_USERAGENT)-1]);
			y += 5;
			useragentno = z++;
		    }
		    x += sizeof(NS_USERAGENT)-1;
		}
		else {
		    strcpy(formatline+y, "%*");
		    y += 2;
		    while (x < length) {
			x++;
			if (line[x] == '%') {
			    x++;
			    break;
			}
		    }
		}
	    }
	    else {
		formatline[y++] = line[x++];
	    }
	}
	formatparms = z-1;
	formatline[y] = 0;

	/* sanity check on the format line */
	if (!ipno || !sysdateno || !requestno || !statusno || !contentlengthno)
	     fprintf(stderr, "Warning: essential information required to produce an ELF logfile is\nmissing from your logfile.\n");

	return(0);
    }

    /* allocate some space for the strings to be stored in */
    x = 9;
    while (x >= 0) {
	if (!(vp[x]=(char *)malloc(strlen(line)+1))) {
	    fprintf(stderr, "Memory allocation error\n");
	    exit(1);
	}
	vp[x][0] = 0;
	x--;
    }

    /* do the scan */
    formatparms = sscanf(line, formatline, vp[1], vp[2], vp[3], vp[4], vp[5], vp[6], vp[7], vp[8], vp[9]);
/*
printf("line: %s\nformatline: %s\n", line, formatline);
printf(" ip: %s\n authuser: %s\n date: %s\n req: %s\n status: %s\n cl: %s\n referrer: %s\n useragent: %s\n", vp[1],vp[2],vp[3],vp[4],vp[5],vp[6],vp[7],vp[8]);
exit(2);
*/
    /* is everything we need here? */
    if (!remoteuserno)
	remoteuser = "-";
    else
	remoteuser = vp[remoteuserno];

    if (!authuserno)
	authuser = "-";
    else
	authuser = vp[authuserno];

    if ( (!ipno || !sysdateno || !requestno || !statusno || !contentlengthno) || (!vp[ipno][0] || !vp[sysdateno][0] || !vp[requestno][0] || !vp[statusno][0] || !vp[contentlengthno][0]) )
	bypass = 1;


    /* output log line */
    if (!bypass || force) {
	fprintf(fileout, "%s %s %s [%s] \"%s\" %s %s", vp[ipno], remoteuser, authuser, vp[sysdateno], vp[requestno], vp[statusno], vp[contentlengthno]);
	if (referrerno && useragentno) fprintf(fileout, " \"%s\" \"%s\"", vp[referrerno], vp[useragentno]);
	fprintf(fileout,"\n");
    }

    /* free the space for the strings to be stored in */
    x = 9;
    while (x >= 0) {
	free(vp[x]);
	x--;
    }

    return(bypass);
}
